import copy
import logging
from typing import TYPE_CHECKING, Any, MutableMapping, MutableSequence, Union

from bson import ObjectId

from nawah.config import Config

if TYPE_CHECKING:
    from nawah.types import NawahDoc, NawahEnv, ResultsArgs

logger = logging.getLogger('nawah')


async def update(
    *,
    env: 'NawahEnv',
    collection_name: str,
    docs: MutableSequence[Union[str, ObjectId]],
    doc: 'NawahDoc',
) -> 'ResultsArgs':
    # Recreate docs list by converting all docs items to ObjectId
    docs = [ObjectId(doc) for doc in docs]
    # Perform update query on matching docs
    collection = env['conn'][Config.data_name][collection_name]
    results = None
    doc = copy.deepcopy(doc)

    # [TODO] Abstract $set pipeline with colon support for all stages

    # Prepare empty update pipeline
    update_pipeline = []

    # Iterate over attrs in doc to set update stage in pipeline
    for attr in doc.keys():
        # Prepare stage pipeline
        update_pipeline_stage_root: MutableMapping[str, Any] = {'$set': {}}
        update_pipeline_stage_current = update_pipeline_stage_root['$set']
        attr_path_part = attr
        attr_path_current = []

        if ':' in attr:
            attr_path = attr.split('.')
            for i, attr_path_part in enumerate(attr_path):
                if i == 0:
                    # First item has to have $
                    attr_path_current.append('$' + attr_path_part.split(':')[0])
                else:
                    attr_path_current.append(attr_path_part.split(':')[0])

                if ':' not in attr_path_part:
                    part_pipeline: MutableMapping[str, Any] = {
                        '$arrayToObject': {
                            '$concatArrays': [
                                {
                                    '$objectToArray': '.'.join(attr_path_current[:-1]),
                                },
                                [
                                    {
                                        'k': attr_path_part,
                                        'v': None,
                                    }
                                ],
                            ]
                        }
                    }
                    if 'v' in update_pipeline_stage_current:
                        update_pipeline_stage_current['v'] = part_pipeline
                    elif 'then' in update_pipeline_stage_current:
                        update_pipeline_stage_current['then'] = part_pipeline
                    else:
                        update_pipeline_stage_current[attr_path_part] = part_pipeline

                    update_pipeline_stage_current = part_pipeline['$arrayToObject'][
                        '$concatArrays'
                    ][1][0]
                else:
                    part_pipeline = {
                        '$map': {
                            'input': '.'.join(attr_path_current),
                            'as': f'this_{i}',
                            'in': {
                                '$cond': {
                                    'if': {
                                        '$eq': [
                                            {
                                                '$indexOfArray': [
                                                    '.'.join(attr_path_current),
                                                    f'$$this_{i}',
                                                ]
                                            },
                                            int(attr_path_part.split(':')[1]),
                                        ]
                                    },
                                    'then': None,
                                    'else': f'$$this_{i}',
                                }
                            },
                        }
                    }

                    if i != 0:
                        # For all subsequent array objects, wrap in object-to-array-to-object pipeline
                        part_pipeline = {
                            '$arrayToObject': {
                                '$concatArrays': [
                                    {
                                        '$objectToArray': '.'.join(
                                            attr_path_current[:-1]
                                        ),
                                    },
                                    [
                                        {
                                            'k': attr_path_part.split(':')[0],
                                            'v': part_pipeline,
                                        }
                                    ],
                                ]
                            }
                        }

                    if 'v' in update_pipeline_stage_current:
                        update_pipeline_stage_current['v'] = part_pipeline
                    elif 'then' in update_pipeline_stage_current:
                        update_pipeline_stage_current['then'] = part_pipeline
                    else:
                        update_pipeline_stage_current[
                            attr_path_part.split(':')[0]
                        ] = part_pipeline

                    if i == 0:
                        update_pipeline_stage_current = part_pipeline['$map']['in'][
                            '$cond'
                        ]
                    else:
                        update_pipeline_stage_current = part_pipeline['$arrayToObject'][
                            '$concatArrays'
                        ][1][0]['v']['$map']['in']['$cond']

                    attr_path_current = [f'$$this_{i}']

        # Check for $add Doc Oper
        if isinstance(doc[attr], dict) and '$add' in doc[attr]:
            add_field = (
                f'${doc[attr]["$field"]}'
                if '$field' in doc[attr].keys() and doc[attr]['$field']
                else f'${".".join(attr_path_current + [attr])}'
            )

            part_pipeline = {
                '$add': [
                    {
                        '$cond': {
                            'if': {'$not': [add_field]},
                            'then': 0,
                            'else': add_field,
                        }
                    },
                    doc[attr]['$add'],
                ]
            }

            # Add part_pipeline to update_pipeline_stage_current
            if 'v' in update_pipeline_stage_current.keys():
                update_pipeline_stage_current['v'] = part_pipeline
            elif 'then' in update_pipeline_stage_current.keys():
                update_pipeline_stage_current['then'] = part_pipeline
            else:
                update_pipeline_stage_current[attr_path_part] = part_pipeline

        # Check for $add Doc Oper
        elif isinstance(doc[attr], dict) and '$multiply' in doc[attr].keys():
            multiply_field = (
                f'${doc[attr]["$field"]}'
                if '$field' in doc[attr].keys() and doc[attr]['$field']
                else f'${".".join(attr_path_current + [attr])}'
            )

            part_pipeline = {
                '$multiply': [
                    {
                        '$cond': {
                            'if': {'$not': [multiply_field]},
                            'then': 0,
                            'else': multiply_field,
                        }
                    },
                    doc[attr]['$multiply'],
                ]
            }

            # Add part_pipeline to update_pipeline_stage_current
            if 'v' in update_pipeline_stage_current.keys():
                update_pipeline_stage_current['v'] = part_pipeline
            elif 'then' in update_pipeline_stage_current.keys():
                update_pipeline_stage_current['then'] = part_pipeline
            else:
                update_pipeline_stage_current[attr_path_part] = part_pipeline

        # Check for $append Doc Oper
        elif isinstance(doc[attr], dict) and '$append' in doc[attr].keys():
            if '$unique' not in doc[attr].keys() or doc[attr]['$unique'] is False:
                part_pipeline = {'$concatArrays': [f'${attr}', [doc[attr]['$append']]]}
            else:
                part_pipeline = {
                    '$concatArrays': [
                        f'${attr}',
                        {
                            '$cond': {
                                'if': {'$in': [doc[attr]['$append'], f'${attr}']},
                                'then': [],
                                'else': [doc[attr]['$append']],
                            }
                        },
                    ]
                }

            # Add part_pipeline to update_pipeline_stage_current
            if 'v' in update_pipeline_stage_current.keys():
                update_pipeline_stage_current['v'] = part_pipeline
            elif 'then' in update_pipeline_stage_current.keys():
                update_pipeline_stage_current['then'] = part_pipeline
            else:
                update_pipeline_stage_current[attr_path_part] = part_pipeline

        # Check for $set_index Doc Oper
        elif isinstance(doc[attr], dict) and '$set_index' in doc[attr]:
            part_pipeline = {
                '$reduce': {
                    'input': f'${attr}',
                    'initialValue': [],
                    'in': {
                        '$concatArrays': [
                            '$$value',
                            {
                                '$cond': {
                                    'if': {
                                        '$eq': [
                                            ['$$this'],
                                            [
                                                {
                                                    '$arrayElemAt': [
                                                        f'${attr}',
                                                        doc[attr]['$index'],
                                                    ]
                                                }
                                            ],
                                        ]
                                    },
                                    'then': [doc[attr]['$set_index']],
                                    'else': ['$$this'],
                                }
                            },
                        ]
                    },
                }
            }

            # Add part_pipeline to update_pipeline_stage_current
            if 'v' in update_pipeline_stage_current.keys():
                update_pipeline_stage_current['v'] = part_pipeline
            elif 'then' in update_pipeline_stage_current.keys():
                update_pipeline_stage_current['then'] = part_pipeline
            else:
                update_pipeline_stage_current[attr_path_part] = part_pipeline

        # Check for $del_val Doc Oper
        elif isinstance(doc[attr], dict) and '$del_val' in doc[attr]:
            part_pipeline = {
                '$reduce': {
                    'input': f'${attr}',
                    'initialValue': [],
                    'in': {
                        '$concatArrays': [
                            '$$value',
                            {
                                '$cond': {
                                    'if': {
                                        '$eq': [
                                            ['$$this'],
                                            doc[attr]['$del_val'],
                                        ]
                                    },
                                    'then': [],
                                    'else': ['$$this'],
                                }
                            },
                        ]
                    },
                }
            }

            # Add part_pipeline to update_pipeline_stage_current
            if 'v' in update_pipeline_stage_current.keys():
                update_pipeline_stage_current['v'] = part_pipeline
            elif 'then' in update_pipeline_stage_current.keys():
                update_pipeline_stage_current['then'] = part_pipeline
            else:
                update_pipeline_stage_current[attr_path_part] = part_pipeline

        # Check for $del_index Doc Oper
        elif isinstance(doc[attr], dict) and '$del_index' in doc[attr]:
            part_pipeline = {
                '$arrayToObject': {
                    '$reduce': {
                        'input': {
                            '$objectToArray': f'${attr}',
                        },
                        'initialValue': [],
                        'in': {
                            '$concatArrays': [
                                '$$value',
                                {
                                    '$cond': {
                                        'if': {
                                            '$eq': [
                                                '$$this.k',
                                                doc[attr]['$del_index'],
                                            ],
                                        },
                                        'then': [],
                                        'else': ['$$this'],
                                    }
                                },
                            ]
                        },
                    }
                }
            }

            # Add part_pipeline to update_pipeline_stage_current
            if 'v' in update_pipeline_stage_current.keys():
                update_pipeline_stage_current['v'] = part_pipeline
            elif 'then' in update_pipeline_stage_current.keys():
                update_pipeline_stage_current['then'] = part_pipeline
            else:
                update_pipeline_stage_current[attr_path_part] = part_pipeline

        else:
            # Add part_pipeline to update_pipeline_stage_current
            if 'v' in update_pipeline_stage_current.keys():
                update_pipeline_stage_current['v'] = {'$literal': doc[attr]}
            elif 'then' in update_pipeline_stage_current.keys():
                update_pipeline_stage_current['then'] = {'$literal': doc[attr]}
            else:
                update_pipeline_stage_current[attr_path_part] = {'$literal': doc[attr]}

        # Add stage to pipeline
        update_pipeline.append(update_pipeline_stage_root)

    logger.debug('Final update query: %s', {'_id': {'$in': docs}})
    logger.debug('Final update pipeline: %s', update_pipeline)

    results = await collection.update_many({'_id': {'$in': docs}}, update_pipeline)
    update_count = results.modified_count

    return {'count': update_count, 'docs': [{'_id': doc} for doc in docs]}



'''Provides 'Config', 'ConfigAttrSys' static classes'''

from typing import (TYPE_CHECKING, Any, Callable, MutableMapping,
                    MutableSequence, Optional, Set, Type, TypedDict)

from nawah.enums import LocaleStrategy

if TYPE_CHECKING:
    from datetime import datetime

    import redis.commands.json.JSON
    from bson import ObjectId
    from motor.motor_asyncio import AsyncIOMotorClient

    from nawah.classes import (L10N, Attr, ClientApp, Job, Module, SysDoc,
                               UserSetting)
    from nawah.types import (AnalyticsEvents, AppPackage, IPQuota, NawahDoc,
                             NawahEnv)


class VarTypeDict(TypedDict):
    '''Provides type-hint for item of 'vars_types' runtime config'''

    package: str
    type: 'Attr'


class ConfigAttrSys:
    '''ConfigAttrSys serves role of static class used as value for 'sys' attribute
    of 'Config' static class'''

    # pylint: disable=too-few-public-methods

    admin_check: bool = False

    conn: 'AsyncIOMotorClient'
    cache: Optional['redis.commands.json.JSON'] = None
    env: 'NawahEnv'
    docs: MutableMapping['ObjectId', 'SysDoc'] = {}
    jobs_base: 'datetime'

    nawah_version: str

    name: str
    version: str
    path: str
    packages: MutableMapping[str, 'AppPackage'] = {}

    cache_channels: Set[str] = set()

    sessions: MutableMapping[str, 'NawahEnv'] = {}
    session_counter = 0
    session_lock: Optional[float] = None
    ip_quota: MutableMapping[str, 'IPQuota'] = {}

    type_attrs: MutableSequence['Attr'] = []


class Config:
    '''Config serves role of static class used as central reference for runtime
    config of Nawah App'''

    # pylint: disable=too-few-public-methods

    debug: bool = False
    env: str
    port: int = 8081

    sys: Type[ConfigAttrSys] = ConfigAttrSys

    test: bool = False

    emulate_test: bool = False

    vars_types: MutableMapping[str, 'VarTypeDict'] = {}
    vars: MutableMapping[str, Any] = {}

    client_apps: MutableMapping[str, 'ClientApp'] = {}

    analytics_events: 'AnalyticsEvents' = {
        'app_conn_verified': True,
        'session_conn_auth': True,
        'session_user_auth': True,
        'session_conn_reauth': True,
        'session_user_reauth': True,
        'session_conn_deauth': True,
        'session_user_deauth': True,
    }

    conn_timeout: int = 120
    quota_anon_min: int = 40
    quota_auth_min: int = 100
    quota_ip_min: int = 500
    file_upload_limit: int = -1
    file_upload_timeout: int = 300

    data_server: str = 'mongodb://localhost'
    data_name: str = 'nawah_data'
    data_ssl: bool = False
    data_disk_use: bool = False

    cache_server: Optional[str] = None
    cache_db: Optional[int] = 0
    cache_username: Optional[str] = None
    cache_password: Optional[str] = None
    cache_expiry: Optional[int] = None

    error_reporting_server: Optional[str] = None

    locales: MutableSequence[str] = ['ar_AE', 'en_GB']
    locale: str = 'ar_AE'
    locale_strategy: 'LocaleStrategy' = LocaleStrategy.DUPLICATE

    admin_doc: 'NawahDoc' = {}
    admin_password: str

    anon_token: str
    anon_privileges: MutableMapping[str, MutableSequence[str]] = {}
    anon_doc: Optional['NawahDoc'] = None

    user_attrs: MutableMapping[str, 'Attr'] = {}
    user_settings: MutableMapping[str, 'UserSetting'] = {}
    user_doc_settings: MutableSequence[str] = []

    groups: MutableSequence[MutableMapping[str, Any]] = []
    default_privileges: MutableMapping[str, MutableSequence[str]] = {}

    data_indexes: MutableSequence[MutableMapping[str, Any]] = []

    docs: MutableSequence['SysDoc'] = []

    l10n: MutableMapping[str, 'L10N'] = {}

    jobs: MutableMapping[str, 'Job'] = {}

    types: MutableMapping[str, Callable] = {}

    modules: MutableMapping[str, 'Module'] = {}



'''Provides 'Cache' dataclass'''

from dataclasses import dataclass
from typing import Callable, Optional

from nawah.config import Config


@dataclass(kw_only=True)
class Cache:
    '''Cache dataclass serves role of defining Cache Instruction, which instructs Nawah of when to
    cache call results. callable for 'condition' should accept any number of following args only:
    'skip_events', 'env', 'query' '''

    channel: str
    condition: Callable
    user_scoped: bool
    period: Optional[int] = None
    file: Optional[bool] = False

    def __post_init__(self):
        Config.sys.cache_channels.add(self.channel)



'''Shared items required by multiple Utilities'''

from math import inf
from typing import Literal, MutableMapping, MutableSequence, TypedDict


class AttrTypeArgValidation(TypedDict, total=False):
    '''Provides type-hint for 'ATTRS_TYPES_ARGS' constant'''

    type: Literal[
        'Attr', 'bool', 'callable', 'dict', 'float', 'int', 'list', 'str', 'union'
    ]
    str_pattern: str
    list_len_range: MutableSequence[float]
    list_item_type: 'AttrTypeArgValidation'  # type: ignore
    dict_key_type: 'AttrTypeArgValidation'  # type: ignore
    dict_val_type: 'AttrTypeArgValidation'  # type: ignore
    union_types: MutableSequence['AttrTypeArgValidation']  # type: ignore
    required: Literal[True]


# Reference to all Attr Types and bound Attr Type Args
ATTRS_TYPES_ARGS: MutableMapping[str, MutableMapping[str, AttrTypeArgValidation]] = {
    'ANY': {},
    'ATTR': {
        'types': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {'type': 'str'},
        }
    },
    'ID': {},
    'STR': {'pattern': {'type': 'str'}},
    'INT': {
        'ranges': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {
                'type': 'list',
                'list_len_range': [2, 3],
                'list_item_type': {'type': 'float'},
            },
        }
    },
    'FLOAT': {
        'ranges': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {
                'type': 'list',
                'list_len_range': [2, 3],
                'list_item_type': {'type': 'float'},
            },
        }
    },
    'BOOL': {},
    'LOCALE': {},
    'LOCALES': {},
    'EMAIL': {
        'allowed_domains': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {'type': 'str'},
        },
        'disallowed_domains': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {'type': 'str'},
        },
        'strict_matching': {'type': 'bool'},
    },
    'PHONE': {
        'codes': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {'type': 'str', 'str_pattern': '[0-9]+'},
        }
    },
    'IP': {},
    'URI_WEB': {
        'allowed_domains': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {'type': 'str'},
        },
        'disallowed_domains': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {'type': 'str'},
        },
        'strict_matching': {'type': 'bool'},
    },
    'DATETIME': {
        'ranges': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {
                'type': 'list',
                'list_len_range': [2, 3],
                'list_item_type': {
                    'type': 'str',
                    'str_pattern': r'[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}(:[0-9]{2}(\.[0-9]{6})?)?|[\-\+][0-9]+[dsmhw]',
                },
            },
        }
    },
    'DATE': {
        'ranges': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {
                'type': 'list',
                'list_len_range': [2, 3],
                'list_item_type': {
                    'type': 'str',
                    'str_pattern': r'[0-9]{4}-[0-9]{2}-[0-9]{2}|[\-\+][0-9]+[dsmhw]',
                },
            },
        }
    },
    'TIME': {
        'ranges': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {
                'type': 'list',
                'list_len_range': [2, 3],
                'list_item_type': {
                    'type': 'str',
                    'str_pattern': r'[0-9]{2}:[0-9]{2}(:[0-9]{2}(\.[0-9]{6})?)?|[\-\+][0-9]+[dsmhw]',
                },
            },
        }
    },
    'FILE': {
        'types': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {'type': 'str'},
        },
        'ratio_ranges': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {
                'type': 'list',
                'list_len_range': [2, 3],
                'list_item_type': {'type': 'float'},
            },
        },
        'dims_ranges': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {
                'type': 'list',
                'list_len_range': [2, 3],
                'list_item_type': {'type': 'float'},
            },
        },
        'size_ranges': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {
                'type': 'list',
                'list_len_range': [2, 3],
                'list_item_type': {'type': 'float'},
            },
        },
    },
    'GEO_POINT': {},
    'LIST': {
        'list': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {'type': 'Attr'},
            'required': True,
        },
        'len_range': {
            'type': 'list',
            'list_len_range': [2, 3],
            'list_item_type': {'type': 'float'},
        },
    },
    'KV_DICT': {
        'key': {
            'type': 'Attr',
            'required': True,
        },
        'val': {
            'type': 'Attr',
            'required': True,
        },
        'len_range': {
            'type': 'list',
            'list_len_range': [2, 3],
            'list_item_type': {'type': 'float'},
        },
        'req': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {'type': 'str'},
        },
    },
    'TYPED_DICT': {
        'dict': {
            'type': 'dict',
            'dict_key_type': {'type': 'str'},
            'dict_val_type': {'type': 'Attr'},
            'required': True,
        }
    },
    'LITERAL': {
        'literal': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {
                'type': 'union',
                'union_types': [
                    {'type': 'str'},
                    {'type': 'int'},
                    {'type': 'float'},
                    {'type': 'bool'},
                ],
            },
            'required': True,
        }
    },
    'UNION': {
        'union': {
            'type': 'list',
            'list_len_range': [1, inf],
            'list_item_type': {'type': 'Attr'},
            'required': True,
        }
    },
    'TYPE': {
        'type': {
            'type': 'union',
            'union_types': [{'type': 'str'}, {'type': 'callable'}],
            'required': True,
        }
    },
}



'''Provides 'file' Module Functions callables'''

from typing import TYPE_CHECKING

from nawah.config import Config
from nawah.enums import AttrType, Event
from nawah.exceptions import FuncException, InvalidAttrException
from nawah.utils import call, extract_attr, validate_attr

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahDoc, NawahEnv, NawahEvents, Results


async def _read(
    skip_events: 'NawahEvents', env: 'NawahEnv', query: 'Query'
) -> 'Results':
    skip_events.append(Event.PERM)
    read_results = await call(
        'base/read', module_name='file', skip_events=skip_events, env=env, query=query
    )

    if read_results['status'] != 200:
        return read_results

    for doc in read_results['args']['docs']:
        doc['file']['lastModified'] = int(doc['file']['lastModified'])

    return read_results


async def _create(env: 'NawahEnv', doc: 'NawahDoc') -> 'Results':
    if Config.file_upload_limit != -1 and len(doc['file']) > Config.file_upload_limit:
        raise FuncException(
            status=400,
            msg='File size is beyond allowed limit',
            args={
                'code': 'INVALID_SIZE',
                'attr': doc['__attr'].decode('utf-8'),
                'name': doc['name'].decode('utf-8'),
            },
        )
    if (module := doc['__module'].decode('utf-8')) not in Config.modules:
        raise FuncException(
            status=400,
            msg=f'Invalid module \'{module}\'',
            args={'code': 'INVALID_MODULE'},
        )

    try:
        attr_type = extract_attr(
            attrs=Config.modules[module].attrs,
            path=(attr := doc['__attr'].decode('utf-8')),
        )
    except (KeyError, ValueError) as e:
        raise FuncException(
            status=400,
            msg=f'Invalid attr \'{attr}\' of module \'{module}\'',
            args={'code': 'INVALID_ATTR'},
        ) from e

    doc = {
        'file': {
            'name': doc['name'].decode('utf-8'),
            'type': doc['type'].decode('utf-8'),
            'size': len(doc['file']),
            'lastModified': int(doc['lastModified'].decode('utf-8')),
            'content': doc['file'],
        },
    }
    try:
        attr_val = doc['file']
        if attr_type.type == AttrType.LIST:
            attr_val = [doc['file']]
        validate_attr(
            mode='create',
            attr_name=attr,
            attr_type=attr_type,
            attr_val=attr_val,
        )
    except InvalidAttrException as e:
        raise FuncException(
            status=400,
            msg=f'Invalid file for \'{attr}\' of module \'{module}\'',
            args={'code': 'INVALID_FILE'},
        ) from e

    create_results = await call(
        'base/create', skip_events=[Event.PERM], module_name='file', env=env, doc=doc
    )

    return create_results



import copy
import logging
from typing import (TYPE_CHECKING, Any, MutableMapping, MutableSequence,
                    Optional, Union, cast)

from bson import ObjectId
from motor.motor_asyncio import AsyncIOMotorCollection

from nawah.classes import Var
from nawah.config import Config
from nawah.enums import AttrType, Event, LocaleStrategy

from ._query import _compile_query

if TYPE_CHECKING:
    from nawah.classes import Attr, Extn, Query
    from nawah.types import NawahDoc, NawahEnv, NawahQuery

logger = logging.getLogger('nawah')


async def read(
    *,
    env: 'NawahEnv',
    collection_name: str,
    attrs: MutableMapping[str, 'Attr'],
    query: 'Query',
    skip_process: bool = False,
    extn_attrs: MutableSequence[str] = [],
) -> MutableMapping[str, Any]:
    skip, limit, sort, group, aggregate_query = _compile_query(
        collection_name=collection_name, attrs=attrs, query=query
    )

    logger.debug(f'aggregate_query: {aggregate_query}')
    logger.debug(f'skip, limit, sort, group: {skip}, {limit}, {sort}, {group}.')

    collection: AsyncIOMotorCollection = env['conn'][Config.data_name][collection_name]
    docs_total_results = collection.aggregate(
        aggregate_query + [{'$count': '__docs_total'}],
        allowDiskUse=Config.data_disk_use,
    )
    try:
        async for doc in docs_total_results:
            docs_total = doc['__docs_total']
        docs_total
    except:
        return {'total': 0, 'count': 0, 'docs': [], 'groups': []}

    groups = {}
    if group:
        for group_condition in group:
            group_query = aggregate_query + [
                {
                    '$bucketAuto': {
                        'groupBy': '$' + group_condition['by'],
                        'buckets': group_condition['count'],
                    }
                }
            ]
            check_group = False
            for i in range(len(group_query)):
                if (
                    list(group_query[i].keys())[0] == '$match'
                    and list(group_query[i]['$match'].keys())[0]
                    == group_condition['by']
                ):
                    check_group = True
                    break
            if check_group:
                del group_query[i]
            group_query_results = collection.aggregate(
                group_query, allowDiskUse=Config.data_disk_use
            )
            groups[group_condition['by']] = [
                {
                    'min': group['_id']['min'],
                    'max': group['_id']['max'],
                    'count': group['count'],
                }
                async for group in group_query_results
            ]

    if sort is not None:
        aggregate_query.append({'$sort': sort})
    if skip is not None:
        aggregate_query.append({'$skip': skip})
    if limit is not None:
        aggregate_query.append({'$limit': limit})

    logger.debug(f'final query: {collection}, {aggregate_query}.')

    docs_count_results = collection.aggregate(
        aggregate_query + [{'$count': '__docs_count'}],
        allowDiskUse=Config.data_disk_use,
    )
    try:
        async for doc in docs_count_results:
            docs_count = doc['__docs_count']
        docs_count
    except:
        return {
            'total': docs_total,
            'count': 0,
            'docs': [],
            'groups': {} if not group else groups,
        }
    docs = collection.aggregate(aggregate_query, allowDiskUse=Config.data_disk_use)
    models = []
    extn_models: MutableMapping[str, Optional[MutableMapping]] = {}
    async for doc in docs:
        if not skip_process:
            doc = await _process_results_doc(
                env=env,
                attrs=attrs,
                doc=doc,
                extn_models=extn_models,
                extn_attrs=extn_attrs,
            )
        if doc:
            models.append(doc)
    return {
        'total': docs_total,
        'count': docs_count,
        'docs': models,
        'groups': {} if not group else groups,
    }


async def _process_results_doc(
    *,
    env: 'NawahEnv',
    attrs: MutableMapping[str, 'Attr'],
    doc: 'NawahDoc',
    extn_models: MutableMapping[str, Optional[MutableMapping]],
    extn_attrs: MutableSequence[str],
) -> MutableMapping[str, Any]:
    # Process doc attrs
    for attr in attrs.keys():
        if attrs[attr].type == AttrType.LOCALE:
            if (
                attr in doc.keys()
                and isinstance(doc[attr], dict)
                and Config.locale in doc[attr].keys()
            ):
                if Config.locale_strategy == LocaleStrategy.NONE_VALUE:
                    doc[attr] = {
                        locale: doc[attr][locale]
                        if locale in doc[attr].keys()
                        else None
                        for locale in Config.locales
                    }
                elif callable(Config.locale_strategy):
                    doc[attr] = {
                        locale: doc[attr][locale]
                        if locale in doc[attr].keys()
                        else Config.locale_strategy(
                            attr_val=doc[attr][Config.locale], locale=locale
                        )
                        for locale in Config.locales
                    }
                else:
                    doc[attr] = {
                        locale: doc[attr][locale]
                        if locale in doc[attr].keys()
                        else doc[attr][Config.locale]
                        for locale in Config.locales
                    }

        if attr not in extn_attrs:
            continue

        await _extend_attr(
            doc=doc,
            scope=doc,
            attr_name=attr,
            attr_type=attrs[attr],
            env=env,
            extn_models=extn_models,
        )
    # Attempt to extned the doc per extns
    return doc


async def _extend_attr(
    *,
    doc: 'NawahDoc',
    scope: Union[MutableMapping[str, Any], MutableSequence[Any]],
    attr_name: Union[str, int],
    attr_type: 'Attr',
    env: 'NawahEnv',
    extn_models: MutableMapping[str, Optional[MutableMapping]],
):
    # If scope is missing attr_name skip
    if isinstance(scope, dict) and attr_name not in scope:
        return

    # Check attr_type for possible types that require deep checking for extending
    if attr_type.type == AttrType.KV_DICT:
        attr_name = cast(str, attr_name)
        scope = cast(MutableMapping[str, Any], scope)
        if scope[attr_name] and isinstance(scope[attr_name], dict):
            for child_attr in scope[attr_name]:
                # attr_type is KV_DICT where Attr Type Arg val could be extended
                await _extend_attr(
                    doc=doc,
                    scope=scope[attr_name],
                    attr_name=child_attr,
                    attr_type=attr_type.args['val'],
                    env=env,
                    extn_models=extn_models,
                )
    if attr_type.type == AttrType.TYPED_DICT:
        attr_name = cast(str, attr_name)
        scope = cast(MutableMapping[str, Any], scope)
        if scope[attr_name] and isinstance(scope[attr_name], dict):
            for child_attr in attr_type.args['dict']:
                # attr_type is TYPED_DICT where each dict item could be extended
                await _extend_attr(
                    doc=doc,
                    scope=scope[attr_name],
                    attr_name=child_attr,
                    attr_type=attr_type.args['dict'][child_attr],
                    env=env,
                    extn_models=extn_models,
                )

    elif attr_type.type == AttrType.LIST:
        attr_name = cast(str, attr_name)
        scope = cast(MutableMapping[str, Any], scope)
        if scope[attr_name] and isinstance(scope[attr_name], list):
            for child_attr in attr_type.args['list']:
                # attr_type is LIST where it could have KV_DICT, TYPED_DICT, ID Attrs Types that can be [deep-]extended
                if child_attr.type == AttrType.KV_DICT:
                    for child_scope in scope[attr_name]:
                        if isinstance(child_scope, dict):
                            for child_child_attr in child_scope:
                                await _extend_attr(
                                    doc=doc,
                                    scope=child_scope,
                                    attr_name=child_child_attr,
                                    attr_type=child_attr.args['val'],
                                    env=env,
                                    extn_models=extn_models,
                                )
                elif child_attr.type == AttrType.TYPED_DICT:
                    for child_scope in scope[attr_name]:
                        if isinstance(child_scope, dict):
                            for child_child_attr in child_attr.args['dict']:
                                await _extend_attr(
                                    doc=doc,
                                    scope=child_scope,
                                    attr_name=child_child_attr,
                                    attr_type=child_attr.args['dict'][child_child_attr],
                                    env=env,
                                    extn_models=extn_models,
                                )
                elif child_attr.type == AttrType.ID:
                    for i in range(len(scope[attr_name])):
                        await _extend_attr(
                            doc=doc,
                            scope=scope[attr_name],
                            attr_name=i,
                            attr_type=child_attr,
                            env=env,
                            extn_models=extn_models,
                        )

    if not attr_type.extn:
        return

    attr_name = cast(str, attr_name)
    scope = cast(MutableMapping[str, Any], scope)
    # Attr is having Extn for _extn value, attempt to extend attr based on scope type
    if isinstance(scope[attr_name], str):
        scope[attr_name] = await _extend_doc(
            env=env,
            doc=doc,
            attr=scope[attr_name],
            extn_id=scope[attr_name],
            extn=attr_type.extn,
            extn_models=extn_models,
        )
    elif isinstance(scope[attr_name], list):
        scope[attr_name] = [
            await _extend_doc(
                env=env,
                doc=doc,
                attr=scope[attr_name],
                extn_id=extn_id,
                extn=attr_type.extn,
                extn_models=extn_models,
            )
            for extn_id in scope[attr_name]
        ]


async def _extend_doc(
    *,
    env: 'NawahEnv',
    doc: 'NawahDoc',
    attr: Optional['NawahDoc'],
    extn_id: str,
    extn: 'Extn',
    extn_models: MutableMapping[str, Optional[MutableMapping]],
) -> Optional[MutableMapping]:
    from nawah.utils import call, var_value

    extn_module_name: str
    # Check if extn module is dynamic value
    if isinstance(extn.module, Var):
        extn_module_name = var_value(extn.module, env=env, doc=doc)
    else:
        extn_module_name = extn.module
    extn_module = Config.modules[extn_module_name]
    # Check if extn attr set to fetch all or specific attrs
    if isinstance(extn.attrs, Var):
        extn_attrs = var_value(extn.attrs, env=env, doc=doc)
        if extn_attrs[0] == '*':
            extn_attrs = list(extn_module.attrs)
    elif extn.attrs[0] == '*':
        extn_attrs = list(extn_module.attrs)
    else:
        extn_attrs = copy.deepcopy(extn.attrs)
    # Implicitly add _id key to extn attrs so that we don't delete it in process
    if '_id' not in extn_attrs:
        extn_attrs.append('_id')
    # Set skip events
    skip_events = [Event.PERM]
    # Check value for Extn.skip_events
    if isinstance(extn.skip_events, list):
        skip_events.extend(extn.skip_events)
    elif isinstance(extn.skip_events, Var):
        skip_events.extend(var_value(extn.skip_events, env=env, doc=doc))
    # Check if extn instruction is explicitly requires second-dimension extn.
    extn_extns = []
    extn_force = False
    if isinstance(extn.force, Var):
        extn_force = var_value(extn.force, env=env, doc=doc)
    if extn.force is True:
        extn_force = True
    if extn_force:
        extn_extns = extn_attrs
    extn_query: 'NawahQuery' = [{'_id': ObjectId(extn_id)}, {'$extn': extn_extns}]
    # Check if additional query steps are to be appended
    # Read doc if not in extn_models
    extn_model_key = str(extn_id) + '_extn_' + str(extn_extns)
    if extn_model_key not in extn_models.keys():
        extn_results = await call(
            f'{extn_module_name}/read',
            skip_events=skip_events,
            env=env,
            query=extn_query,
        )
        if extn_results['args']['count']:
            extn_models[extn_model_key] = extn_results['args']['docs'][0]
        else:
            extn_models[extn_model_key] = None
    # Set attr to extn_models doc
    extn_doc = copy.deepcopy(extn_models[extn_model_key])
    # delete all unneeded keys from the resulted doc
    if extn_doc:
        extn_doc = {attr: extn_doc[attr] for attr in extn_attrs if attr in extn_doc}
    return extn_doc



'''Provides function to create CLI for Nawah'''

import argparse

from nawah import __version__

from ._create import create


def cli():
    '''Creates CLI for Nawah'''

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--version',
        help='Print Nawah version and exit',
        action='version',
        version=f'Nawah v{__version__}',
    )

    subparsers = parser.add_subparsers(
        title='Command',
        description='Nawah CLI command to run',
        dest='command',
    )

    parser_create = subparsers.add_parser('create', help='Create new Nawah app')
    parser_create.set_defaults(func=create)
    parser_create.add_argument(
        'component',
        help='Type of component to create',
        choices=['app', 'package', 'module'],
    )

    parser_create.add_argument(
        'name',
        type=str,
        help='Name of component to create',
    )

    parser_create.add_argument(
        'path',
        type=str,
        nargs='?',
        help='Path to create component in [default .]',
        default='.',
    )

    parser_create.add_argument(
        '--standalone',
        help='When creating Nawah package, add this flag to create standalone Python project '
        'for Nawah package. This allows developing Nawah package in isolation, with ability to '
        'publish it as Python package and use it across multiple apps. If you are intending to '
        'use the package in one app only, don\'t use this flag, and point path to Nawah app '
        'folder, to only create project sub-package',
        action='store_true',
    )

    args = parser.parse_args()

    if args.command:
        args.func(args)

    else:
        parser.print_help()



'''Provides 'file' Nawah Core Module'''

from nawah.classes import Attr, Func, Module, Perm

from ._callables import _create, _read

file = Module(
    name='file',
    desc='\'file\' module provides functionality for \'File Upload Workflow\'',
    collection='file_docs',
    attrs={
        'user': Attr.ID(desc='\'_id\' of \'User\' doc file belongs to'),
        'file': Attr.FILE(desc='File object'),
        'create_time': Attr.DATETIME(
            desc='Python \'datetime\' ISO format of the doc creation'
        ),
    },
    funcs={
        'read': Func(
            permissions=[
                Perm(privilege='__sys'),
            ],
            callable=_read,
        ),
        'create': Func(
            permissions=[Perm(privilege='create')], post_func=True, callable=_create
        ),
        'delete': Func(
            permissions=[
                Perm(privilege='__sys'),
            ],
        ),
    },
)



'''Provides 'group' Nawah Core Module'''

from nawah.classes import Attr, Default, Func, Module, Perm, Var
from nawah.config import Config

group = Module(
    name='group',
    desc='\'group\' module provides data type and controller for groups in Nawah eco-system',
    collection='group_docs',
    attrs={
        'user': Attr.ID(desc='\'_id\' of \'User\' doc the doc belongs to'),
        'name': Attr.LOCALE(desc='Name of the groups as \'LOCALE\''),
        'desc': Attr.LOCALE(
            desc='Description of the group as \'LOCALE\'. This can be used for dynamic generated '
            'groups that are meant to be exposed to end-users'
        ),
        'privileges': Attr.KV_DICT(
            desc='Privileges that any user is a member of the group has',
            key=Attr.STR(),
            val=Attr.LIST(list=[Attr.STR()]),
        ),
        'settings': Attr.KV_DICT(
            desc='\'Setting\' docs to be created, or required for members users when added to '
            'the group',
            key=Attr.STR(),
            val=Attr.ANY(),
        ),
        'create_time': Attr.DATETIME(
            desc='Python \'datetime\' ISO format of the doc creation'
        ),
    },
    defaults={
        'desc': Default(value=lambda _: {locale: '' for locale in Config.locales}),
        'privileges': Default(value={}),
        'settings': Default(value={}),
    },
    funcs={
        'read': Func(
            permissions=[
                Perm(privilege='admin'),
            ],
        ),
        'create': Func(permissions=[Perm(privilege='admin')]),
        'update': Func(
            permissions=[
                Perm(privilege='admin', doc_mod={'groups': None}),
                Perm(
                    privilege='update',
                    query_mod={'_id': Var.SESSION('user')},
                    doc_mod={'groups': None, 'privileges': None},
                ),
            ],
            query_attrs={'_id': Attr.ID()},
        ),
        'delete': Func(
            permissions=[
                Perm(privilege='admin'),
                Perm(privilege='delete', query_mod={'_id': Var.SESSION('user')}),
            ],
            query_attrs={'_id': Attr.ID()},
        ),
    },
)



'''Provides 'create_file' Base Function callable'''

import datetime
import logging
from typing import TYPE_CHECKING

import nawah.data as Data
from nawah.classes import Attr
from nawah.config import Config
from nawah.enums import AttrType
from nawah.utils import extract_attr, validate_doc

from ._shared import FILE_ATTRS
from .exceptions import NoDocUpdatedException, UtilityModuleDataCallException

if TYPE_CHECKING:
    from nawah.types import NawahDoc, NawahEnv, Results

logger = logging.getLogger('nawah')


async def create_file(
    *,
    module_name: str,
    env: 'NawahEnv',
    doc: 'NawahDoc',
) -> 'Results':
    '''Creates file doc for a module'''

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name='create_file'
        )
    if '_id' in doc:
        return await set_file(
            module_name=module_name,
            env=env,
            doc=doc,
            raise_no_success=True,
        )

    file_attr = extract_attr(attrs=module.attrs, path=doc['attr'])

    types_pattern = None
    if file_attr.args['types']:
        types_pattern = '|'.join(file_attr.args['types'])
        # Replace astrisks (*) with dot-astrisks to mark it as valid RegExp pattern
        types_pattern = types_pattern.replace('*', '.*')

    file_doc_attrs = {
        **FILE_ATTRS,
        'file': Attr.TYPED_DICT(
            dict={
                'name': Attr.STR(),
                'lastModified': Attr.INT(),
                'type': Attr.STR(pattern=types_pattern),
                'size': Attr.INT(),
                'content': Attr(
                    desc='__sys_attr',
                    type=AttrType.BYTES,
                    args={},
                ),
            }
        ),
    }

    file_doc = {
        'user': env['session']['user']['_id'],
        'doc': '000000000000000000000000',
        'attr': doc['attr'],
        'file': {
            'name': doc['name'],
            'lastModified': doc['lastModified'],
            'type': doc['type'],
            'size': len(doc['content']),
            'content': doc['content'],
        },
        'create_time': datetime.datetime.utcnow().isoformat(),
    }

    validate_doc(mode='create', attrs=file_doc_attrs, doc=file_doc)

    # Execute Data driver create
    results = await Data.create(
        env=env, collection_name=f'{module.collection}__file', doc=file_doc
    )

    return {'status': 200, 'msg': f'Created {results["count"]} files', 'args': results}


async def set_file(
    *,
    module_name: str,
    env: 'NawahEnv',
    doc: 'NawahDoc',
    raise_no_success: bool,
) -> 'Results':
    '''Sets doc value of file created using 'create_file' '''
    module = Config.modules[module_name]

    update_results = await Data.update(
        env=env,
        collection_name=f'{module.collection}__file',
        docs=[doc['_id']],
        doc={'doc': doc['doc']},
    )

    if raise_no_success is True and update_results['count'] == 0:
        raise NoDocUpdatedException(module_name=module_name)

    return {'status': 200, 'msg': '', 'args': update_results}



'''Provides 'UserSetting' dataclass, 'UserSettingDict' TypedDict'''

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, Optional

from nawah.enums import NawahValues

if TYPE_CHECKING:
    from ._attr import Attr


@dataclass(kw_only=True)
class UserSetting:
    '''SysDoc dataclass serves role of defining 'Setting' docs that are
    required for users, as part of 'AppConfig', 'PackageConfig' '''

    type: Literal['user', 'user_sys']
    val_type: 'Attr'
    default: Optional[Any] = NawahValues.NONE_VALUE

    def __post_init__(self):
        pass



'''Provides 'read' Base Function callable'''

import logging
from typing import TYPE_CHECKING, MutableSequence, Optional

import nawah.data as Data
from nawah.config import Config

from .exceptions import NoDocFoundException, UtilityModuleDataCallException

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahEnv, Results

logger = logging.getLogger('nawah')


async def read(
    *,
    module_name: str,
    env: 'NawahEnv',
    query: 'Query',
    raise_no_success: Optional[bool],
) -> 'Results':
    '''Reads docs from module collection that match query'''

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name='create'
        )

    extn_list: MutableSequence[str] = []

    extn_list = [attr_name.split('.')[0] for attr_name in module.extns]

    if '$extn' in query:
        if query['$extn'] is False:
            extn_list = []
        elif isinstance(query['$extn'], list):
            extn_list = [
                attr_name
                for attr_name in query['$extn']
                if attr_name.split(':')[0].split('.')[0] in extn_list
            ]

    if '$attrs' in query:
        extn_list = [
            attr_name for attr_name in extn_list if attr_name in query['$attrs']
        ]

    results = await Data.read(
        env=env,
        collection_name=module.collection,
        attrs=module.attrs,
        query=query,
        extn_attrs=extn_list,
    )

    # [DOC] if $attrs query arg is present return only required keys.
    if '$attrs' in query:
        query['$attrs'].insert(0, '_id')
        for i in range(len(results['docs'])):
            results['docs'][i] = {
                attr: results['docs'][i][attr]
                for attr in query['$attrs']
                if attr in results['docs'][i]
            }

    if raise_no_success is True and results['count'] == 0:
        raise NoDocFoundException(module_name=module_name)

    return {'status': 200, 'msg': f'Found {results["count"]} docs', 'args': results}



'''Provides '''

import datetime
from typing import TYPE_CHECKING, Literal, MutableMapping, MutableSequence

from bson import ObjectId

from nawah.config import Config

if TYPE_CHECKING:
    from nawah.types import NawahEnv, NawahSession, Results, ResultsArgs


def mock_results(
    *, status: int, msg: str = None, args: 'ResultsArgs' = None
) -> 'Results':
    '''Returns dict structured as call results'''

    return {
        'status': status,
        'msg': msg or 'Mock results',
        'args': args or {},
    }


def mock_env(*, session: 'NawahSession' = None) -> 'NawahEnv':
    '''Returns dict structured as 'Env' dict. If no value provided for 'session', value would be
    return of 'mock_session()' '''

    return {
        'client_app': '__test',
        'session': session or mock_session(),
        'args': {},
    }


def mock_session(
    *, user: MutableMapping = None, groups: MutableSequence[str] = None
) -> 'NawahSession':
    '''Returns dict structured as 'session' of 'Env' dict. If no value provided for 'user', value
    would be return of 'mock_user()'. If no value provided for 'groups', value would be empty
    list'''

    return {
        'user': user or {},
        'groups': groups or [],
        'host_add': '127.0.0.1',
        'user_agent': 'Nawah Framework Testing Tools',
        'expiry': datetime.datetime.utcnow().isoformat(),
        'token_hash': '__token',
        'create_time': datetime.datetime.utcnow().isoformat(),
    }


def mock_user(  # pylint: disable=too-many-arguments
    *,
    _id: ObjectId = None,
    name: MutableMapping[str, str] = None,
    locale: str = None,
    create_time: str = None,
    login_time: str = None,
    groups: MutableSequence[str] = None,
    privileges: MutableMapping[str, MutableSequence[str]] = None,
    status: Literal['active', 'banned', 'deleted', 'disabled_password'] = None,
    **kwargs
) -> MutableMapping:
    '''Thin wrapper for 'mock_doc'. Returns dict structured as 'user' doc'''

    return mock_doc(
        _id=_id,
        name=name or {locale: '__name' for locale in Config.locales},
        locale=locale or Config.locale,
        create_time=create_time or datetime.datetime.utcnow().isoformat(),
        login_time=login_time or datetime.datetime.utcnow().isoformat(),
        groups=groups or ['f00000000000000000000013'],
        privileges=privileges or {},
        status=status,
        **kwargs,
    )


def mock_doc(_id: ObjectId = None, **kwargs):
    '''Returns dict structured as doc'''

    return {
        '_id': _id or ObjectId(),
        **kwargs,
    }



import logging
import os
import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import argparse

logger = logging.getLogger('nawah')

NAME_PATTERN = r'^[a-z][a-z0-9_]+$'


def create(args: 'argparse.Namespace'):
    '''Provides create command functionality to Nawah CLI'''

    if not re.match(r'^[a-z][a-z0-9_]+$', args.name):
        raise Exception(
            'Value for \'name\' CLI Arg is invalid. Name should have only small letters, '
            'numbers, and underscores.'
        )

    component = args.component
    path = os.path.realpath(args.path)

    logger.info(
        'Checking possibility to create component \'%s\', at \'%s\'', component, path
    )

    if os.path.exists(directory := os.path.join(path, args.name)):
        raise Exception(f'Path \'{path}\' already has item \'{args.name}\'')

    os.mkdir(directory)
    callables = {
        'app': create_app,
    }
    callables[component](directory, args.name, args.standalone)


def create_app(directory: str, name: str, _):
    with open(os.path.join(directory, '__init__.py'), 'w', encoding='UTF-8') as f:
        pass



'''Provides 'update' Base Function callable'''

import inspect
import logging
from typing import TYPE_CHECKING, Any, MutableSequence, Optional, cast

import nawah.data as Data
from nawah.config import Config
from nawah.enums import Event
from nawah.utils import call, validate_doc

from .exceptions import (DuplicateUniqueException, EmptyUpdateDocException,
                         NoDocUpdatedException, UpdateMultiUniqueException,
                         UtilityModuleDataCallException)

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahDoc, NawahEnv, NawahEvents, Results

logger = logging.getLogger('nawah')


async def update(
    *,
    module_name: str,
    skip_events: 'NawahEvents',
    env: 'NawahEnv',
    query: 'Query',
    doc: 'NawahDoc',
    raise_no_success: Optional[bool],
) -> 'Results':
    '''Updates docs for a module matching query 'query' '''

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name='update'
        )

    # Check presence and validate all attrs in doc args
    validate_doc(
        mode='update',
        doc=doc,
        attrs=module.attrs,
    )
    # Delete all attrs not belonging to the module
    shadow_doc = {}
    for attr_name in ['_id', *doc.keys()]:
        attr_root = attr_name.split('.')[0].split(':')[0]
        # Check top-level attr belong to module
        if attr_root not in module.attrs.keys():
            continue
        # Check attr is valid Doc Oper
        if (
            isinstance(doc[attr_name], dict)
            and doc[attr_name].keys()
            and list(doc[attr_name].keys())[0][0] == '$'
            and doc[attr_name][list(doc[attr_name].keys())[0]] is None
        ):
            continue
        # Add non-None attrs to shadow_doc
        if doc[attr_name] is not None:
            shadow_doc[attr_name] = doc[attr_name]

    doc = shadow_doc

    # Check if there is anything yet to update
    if not doc:
        if raise_no_success:
            raise EmptyUpdateDocException(module_name=module_name)

        return {'status': 200, 'msg': 'Nothing to update', 'args': {}}
    # Find which docs are to be updated
    docs_results = await Data.read(
        env=env,
        collection_name=module.collection,
        attrs=module.attrs,
        query=query,
        skip_process=True,
    )
    # Check unique_attrs
    if module.unique_attrs:
        # If any of the unique_attrs is present in doc, and docs_results is > 1, we have duplication
        if len(docs_results['docs']) > 1:
            unique_attrs_check = True
            for attr in module.unique_attrs:
                if isinstance(attr, str) and attr in doc:
                    unique_attrs_check = False
                    break

                if isinstance(attr, tuple):
                    for child_attr in attr:
                        if not unique_attrs_check:
                            break

                        if child_attr in doc:
                            unique_attrs_check = False
                            break

            if not unique_attrs_check:
                raise UpdateMultiUniqueException()

        # Check if any of the unique_attrs are present in doc
        if any(attr in module.unique_attrs for attr in doc):
            # Check if the doc would result in duplication after update
            unique_attrs_query: MutableSequence[Any] = [[]]
            for attr in module.unique_attrs:
                if isinstance(attr, str):
                    attr = cast(str, attr)
                    if attr in doc:
                        unique_attrs_query[0].append({attr: doc[attr]})
                elif isinstance(attr, tuple):
                    unique_attrs_query[0].append(
                        {
                            child_attr: doc[child_attr]
                            for child_attr in attr
                            if attr in doc
                        }
                    )
            unique_attrs_query.append(
                {'_id': {'$nin': [doc['_id'] for doc in docs_results['docs']]}}
            )
            unique_attrs_query.append({'$limit': 1})
            unique_results = await call(
                'base/read',
                module_name=module_name,
                skip_events=[Event.PERM],
                env=env,
                query=unique_attrs_query,
            )
            if unique_results['args']['count']:
                raise DuplicateUniqueException(unique_attrs=module.unique_attrs)

    create_diff = False
    # If module has diff enabled, and Event DIFF not skipped:
    if module.diff and Event.DIFF not in skip_events:
        create_diff = True
        # Attr Type TYPE diff, call the funcion and catch InvalidAttrException
        try:
            func_params = {
                'mode': 'create',
                'attr_name': 'diff',
                'attr_type': module.diff,
                'attr_val': None,
                'skip_events': skip_events,
                'env': env,
                'query': query,
                'doc': doc,
                'scope': doc,
            }
            await module.diff.condition(
                **{
                    param: func_params[param]
                    for param in inspect.signature(module.diff.condition).parameters
                }
            )

        except:
            # [TODO] Implement similar error logging as in retrieve_file
            create_diff = False
            logger.debug('Skipped Diff Workflow due to failed condition')

    else:
        logger.debug(
            'Skipped Diff Workflow due to: %s, %s',
            module.diff,
            Event.DIFF not in skip_events,
        )

    results = await Data.update(
        env=env,
        collection_name=module.collection,
        docs=[doc['_id'] for doc in docs_results['docs']],
        doc=doc,
    )

    if results['count'] and create_diff:
        for doc_result in docs_results['docs']:
            diff_results = await Data.create(
                env=env,
                collection_name=f'{module.collection}__diff',
                doc={
                    'user': env['session']['user']['_id'],
                    'doc': doc_result['_id'],
                    'attrs': {
                        attr.split('.')[0]: doc_result[attr.split('.')[0]]
                        for attr in doc
                    },
                },
            )
        if not diff_results['count']:
            logger.error('Failed to create Diff doc. Results: %s', diff_results)

    # update soft action is to only return the new created doc _id.
    if Event.SOFT in skip_events:
        read_results = await call(
            'base/read',
            module_name=module_name,
            skip_events=[Event.PERM],
            env=env,
            query=[{'_id': {'$in': [doc['_id'] for doc in docs_results['docs']]}}],
        )
        results = read_results['args']

    if raise_no_success is True and results['count'] == 0:
        raise NoDocUpdatedException(module_name=module_name)

    return {'status': 200, 'msg': f'Updated {results["count"]} docs', 'args': results}



'''Provides 'Extn' dataclass'''

from dataclasses import dataclass
from typing import TYPE_CHECKING, MutableSequence, Optional, Union

if TYPE_CHECKING:
    from nawah.classes import Var
    from nawah.enums import Event
    from nawah.types import NawahQuery


@dataclass(kw_only=True)
class Extn:
    '''Extn dataclass serves role of defining Extension Instruction, which
    instructs Nawah to extend specific attr value unto another doc'''

    module: Union[str, 'Var']
    attrs: Union[MutableSequence[str], 'Var']
    force: Union[bool, str, 'Var'] = False
    skip_events: Optional[Union[MutableSequence['Event'], 'Var']] = None
    query: Optional[Union['NawahQuery', 'Var']] = None

    def __post_init__(self):
        pass



'''Provides 'Env', 'Package', 'App' dataclasses'''

from dataclasses import dataclass
from typing import (TYPE_CHECKING, Any, Callable, MutableMapping,
                    MutableSequence, Optional, Union)

if TYPE_CHECKING:
    from nawah.types import AnalyticsEvents, NawahDoc

    from ._attr import Attr
    from ._client_app import ClientApp
    from ._job import Job
    from ._l10n import L10N
    from ._module import Module
    from ._sys_doc import SysDoc
    from ._user_setting import UserSetting
    from ._var import Var


@dataclass(kw_only=True)
class Env:
    '''Env dataclass serves role of defining runtime environment config.
    It is used as item for 'envs' App Config Attr. It also serves as base
    for 'Package' dataclass'''

    # pylint: disable=too-many-instance-attributes

    debug: Union[bool, 'Var'] = False
    port: Optional[Union[int, 'Var']] = None
    emulate_test: bool = False
    desc: Optional[str] = None
    vars: Optional[MutableMapping[str, Any]] = None
    client_apps: Optional[MutableMapping[str, 'ClientApp']] = None
    analytics_events: Optional['AnalyticsEvents'] = None
    conn_timeout: Optional[Union[int, 'Var']] = None
    quota_anon_min: Optional[Union[int, 'Var']] = None
    quota_auth_min: Optional[Union[int, 'Var']] = None
    quota_ip_min: Optional[Union[int, 'Var']] = None
    file_upload_limit: Optional[Union[int, 'Var']] = None
    file_upload_timeout: Optional[Union[int, 'Var']] = None
    data_server: Optional[Union[str, 'Var']] = None
    data_name: Optional[Union[str, 'Var']] = None
    data_ssl: Optional[Union[bool, 'Var']] = None
    data_disk_use: Optional[Union[bool, 'Var']] = None
    cache_server: Optional[Union[str, 'Var']] = None
    cache_db: Optional[Union[int, 'Var']] = 0
    cache_username: Optional[Union[str, 'Var']] = None
    cache_password: Optional[Union[str, 'Var']] = None
    cache_expiry: Optional[Union[int, 'Var']] = None
    error_reporting_server: Optional[Union[str, 'Var']] = None
    locales: Optional[MutableSequence[str]] = None
    locale: Optional[str] = None
    l10n: Optional[MutableMapping[str, 'L10N']] = None
    admin_doc: Optional['NawahDoc'] = None
    admin_password: Optional['Var'] = None
    anon_token: Optional[str] = None
    anon_privileges: Optional[MutableMapping[str, MutableSequence[str]]] = None
    anon_doc: Optional['NawahDoc'] = None
    user_attrs: Optional[MutableMapping[str, 'Attr']] = None
    user_settings: Optional[MutableMapping[str, 'UserSetting']] = None
    user_doc_settings: Optional[MutableSequence[str]] = None
    groups: Optional[MutableSequence[MutableMapping[str, Any]]] = None
    default_privileges: Optional[MutableMapping[str, MutableSequence[str]]] = None
    data_indexes: Optional[MutableSequence[MutableMapping[str, Any]]] = None
    docs: Optional[MutableSequence['SysDoc']] = None
    jobs: Optional[MutableMapping[str, 'Job']] = None
    types: Optional[MutableMapping[str, Callable]] = None


@dataclass(kw_only=True)
class Package(Env):
    '''Package dataclass serves role of defining Nawah Package and its config.
    It also serves as base for 'App' dataclass'''

    # pylint: disable=too-many-instance-attributes

    name: Optional[str] = None
    api_level: Optional[str] = None
    version: Optional[str] = None
    vars_types: Optional[MutableMapping[str, 'Attr']] = None
    modules: Optional[MutableSequence['Module']] = None

    def __post_init__(self):
        if self.name is None:
            raise Exception('Config Attr \'name\' should have value')


@dataclass(kw_only=True)
class App(Package):
    '''App dataclass serves role as defining Nawah App and its config'''

    # pylint: disable=too-many-instance-attributes

    name: Optional[str] = None
    version: Optional[str] = None
    env: Optional[Union[str, 'Var']] = None
    envs: Optional[MutableMapping[str, Env]] = None
    force_admin_check: Optional[bool] = None
    packages: Optional[MutableSequence['Package']] = None

    def __post_init__(self):
        if self.version is None:
            raise Exception('Config Attr \'version\' should have value')



'''Provides 'core' Nawah Package'''

from nawah.classes import Package

from .base import base
from .file import file
from .group import group
from .session import session
from .setting import setting
from .user import user

core = Package(
    name='core',
    api_level='2.0',
    version='2.0.0',
    modules=[base, user, group, session, setting, file],
)



'''Provides 'setting' Module Functions callables'''

from typing import TYPE_CHECKING

from nawah.config import Config
from nawah.enums import Event
from nawah.exceptions import FuncException
from nawah.utils import call, decode_attr_type, validate_doc

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahDoc, NawahEnv, NawahEvents, Results


async def _create(
    env: 'NawahEnv', doc: 'NawahDoc', raise_no_success: bool
) -> 'Results':
    results = await call(
        'base/create',
        skip_events=[Event.PERM],
        module_name='setting',
        env=env,
        doc=doc,
        args={'raise_no_success': raise_no_success},
    )

    if results['status'] != 200:
        return results

    if doc['type'] in ['user', 'user_sys']:
        if (
            doc['user'] == env['session']['user']['_id']
            and doc['var'] in Config.user_doc_settings
        ):
            env['session']['user'][doc['var']] = doc['val']

    return results


async def _update(
    skip_events: 'NawahEvents',
    env: 'NawahEnv',
    query: 'Query',
    doc: 'NawahDoc',
    raise_no_success: bool,
) -> 'Results':
    for attr in doc.keys():
        if attr == 'val' or attr.startswith('val.'):
            break
    else:
        raise FuncException(
            status=400,
            msg='Could not match doc with any of the required doc_args. Failed sets: '
            '[\'val\': Missing]',
            args={'code': 'INVALID_DOC'},
        )

    setting_results = await call(
        'setting/read',
        skip_events=[Event.PERM, *skip_events],
        env=env,
        query=query,
    )
    if not setting_results['args']['count']:
        raise FuncException(
            status=400, msg='Invalid Setting doc', args={'code': 'INVALID_SETTING'}
        )

    setting = setting_results['args']['docs'][0]

    if Event.ATTRS_DOC not in skip_events:
        # [DOC] Attempt to validate val against Setting val_type
        setting_val_type = decode_attr_type(encoded_attr_type=setting['val_type'])
        validate_doc(
            mode='update',
            doc=doc,
            attrs={'val': setting_val_type},
        )

    results = await call(
        'base/update',
        skip_events=[Event.PERM],
        module_name='setting',
        env=env,
        doc=doc,
        args={'raise_no_success': raise_no_success},
    )

    if results['status'] != 200:
        return results

    # Safely check if type in query, as Event ATTRS_QUERY might skipped confirming its existence
    if 'type' in query and query['type'][0] in ['user', 'user_sys']:
        if (
            query['user'][0] == env['session']['user']['_id']
            and query['var'][0] in Config.user_doc_settings
        ):
            env['session']['user'][doc['var']] = doc['val']

    return results



'''Provides 'read' Base Function callable'''

from typing import TYPE_CHECKING

import nawah.data as Data
from nawah.classes import Attr
from nawah.config import Config

from .exceptions import UtilityModuleDataCallException

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahEnv, Results

DIFF_ATTRS = {
    'user': Attr.ID(),
    'doc': Attr.ID(),
    'attrs': Attr.KV_DICT(key=Attr.STR(), val=Attr.ANY()),
}


async def read_diff(
    *,
    module_name,
    env: 'NawahEnv',
    query: 'Query',
) -> 'Results':
    '''Reads diff docs for module'''

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name='read_diff'
        )

    results = await Data.read(
        env=env,
        collection_name=f'{module.collection}__diff',
        attrs=DIFF_ATTRS,
        query=query,
        extn_attrs=[],
    )

    return {'status': 200, 'msg': f'Found {results["count"]} docs', 'args': results}



'''Provides 'Query' and related classes and types'''

import copy
from typing import (TYPE_CHECKING, Any, Literal, MutableMapping,
                    MutableSequence, Optional, TypedDict, Union, cast)

from nawah.exceptions import InvalidQueryArgException, UnknownQueryArgException

if TYPE_CHECKING:
    from nawah.types import NawahQuery, NawahQuerySpecial


SPECIAL_ATTRS = [
    '$search',
    '$sort',
    '$skip',
    '$limit',
    '$extn',
    '$soft',
    '$attrs',
    '$group',
    '$geo_near',
]


class QueryIndexRecord(TypedDict):
    '''Provides type-hint for a single record in Query Index'''

    oper: str
    path: str
    val: Any


class Query(list):
    '''Compiles a list-based Nawah Query into an object that is indexable per \'Query Args\', \'Query Args Opers\'.
    Class provides secure methods to access and manipulate values in Query'''

    _query: 'NawahQuery'
    _special: 'NawahQuerySpecial'
    _index: MutableMapping[str, MutableSequence[QueryIndexRecord]]

    def __init__(self, query: Union['NawahQuery', 'Query']):
        self._query = query
        if isinstance(self._query, Query):
            query = cast(Query, query)
            self._query = query._query + [query._special]
        self._special = {}
        self._index = {}
        self._create_index(self._query)
        super().__init__(self._query)

    def __repr__(self):
        return str(self._query + [self._special])

    def _create_index(self, query: 'NawahQuery', path=None):
        if not path:
            path = []
            self._index = {}
        for i, _ in enumerate(query):
            if isinstance(query[i], dict):
                del_attrs = []
                for attr in query[i].keys():
                    if attr in SPECIAL_ATTRS:
                        self._special[attr] = query[i][attr]  # type: ignore
                        del_attrs.append(attr)
                    elif attr.startswith('__or'):
                        self._create_index(query[i][attr], path=path + [i, attr])  # type: ignore
                    else:
                        if (
                            isinstance(query[i][attr], dict)  # type: ignore
                            and len(query[i][attr].keys()) == 1  # type: ignore
                            and list(query[i][attr].keys())[0][0] == '$'  # type: ignore
                        ):
                            attr_oper = list(query[i][attr].keys())[0]  # type: ignore
                        else:
                            attr_oper = '$eq'
                        if attr not in self._index:
                            self._index[attr] = []
                        if isinstance(query[i][attr], dict) and '_id' in query[i][attr]:  # type: ignore
                            query[i][attr] = query[i][attr]['_id']  # type: ignore
                        validate_query_arg(arg_name=attr, arg_oper=attr_oper, arg_val=query[i][attr])  # type: ignore
                        self._index[attr].append(
                            {
                                'oper': attr_oper,
                                'path': path + [i],
                                'val': query[i][attr],  # type: ignore
                            }
                        )
                for attr in del_attrs:
                    del query[i][attr]  # type: ignore
            elif isinstance(query[i], list):
                self._create_index(query[i], path=path + [i])  # type: ignore
            else:
                raise ValueError(
                    f'Type of Query Arg is of type \'{type(query[i])}\'. Only types \'dict, list\' are allowed.'
                )
        if not path:
            self._query = self._sanitise_query()

    def _sanitise_query(self, query: 'NawahQuery' = None):
        if query is None:
            query = self._query
        query = cast('NawahQuery', query)
        query_shadow = []
        for step in query:
            if isinstance(step, dict):
                for attr in step.keys():
                    if attr.startswith('__or'):
                        step[attr] = self._sanitise_query(step[attr])  # type: ignore
                        if len(step[attr]):  # type: ignore
                            query_shadow.append(step)
                            break
                    elif attr[0] != '$':
                        query_shadow.append(step)
                        break
            elif isinstance(step, list):
                step = self._sanitise_query(step)
                if len(step) != 0:
                    query_shadow.append(step)
        return query_shadow

    def __deepcopy__(self, memo):
        return Query(copy.deepcopy(self._query + [self._special]))

    def append(self, obj: Any):
        self._query.append(obj)
        self._create_index(self._query)
        super().__init__(self._query)

    def __contains__(self, attr):
        if attr in SPECIAL_ATTRS:
            return attr in self._special

        if ':' in attr:
            attr_index, attr_oper = attr.split(':')
        else:
            attr_index = attr
            attr += ':$eq'
            attr_oper = '$eq'

        if attr_index in self._index:
            for val in self._index[attr_index]:
                if attr_oper in (val['oper'], '*'):
                    return True

        return False

    def __getitem__(self, attr):
        if attr in SPECIAL_ATTRS:
            return self._special[attr]

        attrs = []
        vals = []
        paths: MutableSequence[MutableSequence[int]] = []
        indexes = []
        attr_filter: Optional[str] = None
        oper_filter: Optional[str] = None

        if attr.split(':')[0] != '*':
            attr_filter = attr.split(':')[0]

        if ':' not in attr:
            oper_filter = '$eq'
            attr += ':$eq'
        elif ':*' not in attr:
            oper_filter = attr.split(':')[1]

        for index_attr in self._index:
            if attr_filter and index_attr != attr_filter:
                continue
            i = 0
            for val in self._index[index_attr]:
                if not oper_filter or (oper_filter and val['oper'] == oper_filter):
                    attrs.append(index_attr)
                    # [TODO] Simplify this condition by enforcing Query Args with $eq Query Oper are always stripped down to value
                    if not oper_filter or (
                        oper_filter == '$eq'
                        and (
                            not isinstance(val['val'], dict) or '$eq' not in val['val']
                        )
                    ):
                        vals.append(val['val'])
                    else:
                        vals.append(val['val'][oper_filter])
                    paths.append(val['path'])
                    indexes.append(i)
                    i += 1
        return QueryAttrList(
            query=self, attrs=attrs, paths=paths, indexes=indexes, vals=vals
        )

    def __setitem__(self, attr, val):
        if attr[0] != '$':
            raise Exception('Non-special attrs can only be updated by attr index.')
        self._special[attr] = val

    def __delitem__(self, attr):
        if attr[0] != '$':
            raise Exception('Non-special attrs can only be deleted by attr index.')
        del self._special[attr]


class QueryAttrList(list):
    '''A meta class that creates references to specific \'Query Arg\' in a \'Query\'
    object, to allow manipulation or deletion of specific \'Query Arg\' value'''

    def __init__(
        self,
        *,
        query: Query,
        attrs: MutableSequence[str],
        paths: MutableSequence[MutableSequence[int]],
        indexes: MutableSequence[int],
        vals: MutableSequence[Any],
    ):
        self._query = query
        self._attrs = attrs
        self._paths = paths
        self._indexes = indexes
        self._vals = vals
        super().__init__(vals)

    def __setitem__(self, item, val):
        if item == '*':
            for i in range(len(self._vals)):
                self.__setitem__(i, val)
        else:
            instance_attr = self._query._query
            for path_part in self._paths[item]:
                instance_attr = instance_attr[path_part]
            instance_attr[self._attrs[item].split(':')[0]] = val
            self._query._create_index(self._query._query)

    def __delitem__(self, item):
        if item == '*':
            for i in range(len(self._vals)):
                self.__delitem__(i)
        else:
            instance_attr = self._query._query
            for path_part in self._paths[item]:
                instance_attr = instance_attr[path_part]
            del instance_attr[self._attrs[item].split(':')[0]]
            self._query._create_index(self._query._query)

    def replace_attr(self, item: Union[Literal['*'], int], new_attr: str):
        '''Replaces attr name of \'Query Arg\', retaining the value[s] and position in \'Query\' object'''

        if item == '*':
            for i in range(len(self._vals)):
                self.replace_attr(i, new_attr)
            return

        instance_attr = self._query._query
        for path_part in self._paths[item]:
            instance_attr = instance_attr[path_part]  # type: ignore
        # Set new attr
        instance_attr[new_attr] = instance_attr[self._attrs[item].split(':')[0]]  # type: ignore
        # Delete old attr
        del instance_attr[self._attrs[item].split(':')[0]]  # type: ignore
        # Update index
        self._query._create_index(self._query._query)


def validate_query_arg(*, arg_name, arg_oper, arg_val):
    '''Validates value of provided \'Query Arg\' per \'Query Arg Oper\'. If failed, raises
    \'InvalidQueryArgException\' or \'UnknownQueryArgException\' if \'Query Arg Oper\' is unknow'''

    if arg_oper in ['$ne', '$eq']:
        return

    if arg_oper in ['$gt', '$gte', '$lt', '$lte']:
        if type(arg_val[arg_oper]) not in [str, int, float]:
            raise InvalidQueryArgException(
                arg_name=arg_name,
                arg_oper=arg_oper,
                arg_type=[str, int, float],
                arg_val=arg_val[arg_oper],
            )
    elif arg_oper in ['$all', '$in', '$nin']:
        if not isinstance(arg_val[arg_oper], list):
            # [TODO] Validate why this was added
            # or not len(arg_val[arg_oper]):
            raise InvalidQueryArgException(
                arg_name=arg_name,
                arg_oper=arg_oper,
                arg_type=list,
                arg_val=arg_val[arg_oper],
            )
    elif arg_oper == '$regex':
        if isinstance(arg_val[arg_oper], str):
            raise InvalidQueryArgException(
                arg_name=arg_name,
                arg_oper=arg_oper,
                arg_type=str,
                arg_val=arg_val[arg_oper],
            )
    else:
        raise UnknownQueryArgException(arg_name=arg_name, arg_oper=arg_oper)



'''Provides exceptions related to 'var_value' Utility'''


class InvalidLocaleException(Exception):
    '''Raied by 'var_value' Utility if 'locale' is not defined in runtime config'''

    def __init__(self, *, locale):
        super().__init__(
            InvalidLocaleException.format_msg(locale=locale),
            {'locale': locale},
        )

    @staticmethod
    def format_msg(*, locale):
        '''Formats exception message'''
        return f'Invalid locale \'{locale}\''


class InvalidVarException(Exception):
    '''Raied by 'var_value' Utility if failed to extract value of 'var' '''

    def __init__(self, *, var):
        super().__init__(
            InvalidVarException.format_msg(var=var),
            {'var': var},
        )

    @staticmethod
    def format_msg(*, var):
        '''Formats exception message'''
        return f'Invalid var \'{var}\''
